unit SGClass;

interface

uses
  WaveInOut, WaveFmt, GenUtils;

type
  TSineGen = class;      

  PSGWaveState = ^TSGWaveState;
  TSGWaveState = record
    Phase, PhaseInc, Amp_0, Amp_1, PhaseOfs_0, PhaseOfs_1: Real;
  end;

  TSGParamTag = (sgpActive, sgpFrequency, sgpLevel, sgpLevel_0, sgpLevel_1,
    sgpPhaseDifference, sgpPhaseOffset, sgpLatency, sgpPrebuffer, sgpPCMFormat,
    sgpDevice);
  TSGParamMask = set of TSGParamTag;

  TSGParamsEvent = procedure(Sender: TSineGen; ParamTypes: TSGParamMask) of object;

  TSineGen = class(TWaveOut)
  private
    FFrequency: Real;
    FLevel_0: Real;
    FLevel_1: Real;
    FPhaseDifference: Real;
    FPhaseOffset: Real;
    FWaveState: TSGWaveState;
    FOnParamsChange: TSGParamsEvent;
    function GetActive: Boolean;
    function GetDeviceID: Cardinal;
    function GetLatency: Real;
    function GetLevel: Real;
    function GetMaxAvailFreq: Real;
    function GetPCMFormat: TPCMFormat;
    function GetPrebuffer: Real;
    procedure SetActive(const Value: Boolean);
    procedure SetDeviceID(const Value: Cardinal);
    procedure SetFrequency(const Value: Real);
    procedure SetLatency(const Value: Real);
    procedure SetLevel(const Value: Real);
    procedure SetLevel_0(const Value: Real);
    procedure SetLevel_1(const Value: Real);
    procedure SetPCMFormat(const Value: TPCMFormat);
    procedure SetPhaseDifference(const Value: Real);
    procedure SetPhaseOffset(const Value: Real);
    procedure SetPrebuffer(const Value: Real);
  protected
    procedure ParamsChange(ParamMask: TSGParamMask); virtual;
    procedure ProcessData(var Data; SampleCount: Integer); override;
    procedure RestoreWaveState(const Source); override;
    procedure SaveWaveState(var Dest); override;
    function WaveStateDataSize: Integer; override;
  public
    constructor Create; virtual;
    property Active: Boolean read GetActive write SetActive;
    property DeviceID: Cardinal read GetDeviceID write SetDeviceID;
    property Frequency: Real read FFrequency write SetFrequency;
    property Latency: Real read GetLatency write SetLatency;
    property Level: Real read GetLevel write SetLevel;
    property Level_0: Real read FLevel_0 write SetLevel_0;
    property Level_1: Real read FLevel_1 write SetLevel_1;
    property MaxAvailFreq: Real read GetMaxAvailFreq;
    property PCMFormat: TPCMFormat read GetPCMFormat write SetPCMFormat;
    property PhaseDifference: Real read FPhaseDifference write SetPhaseDifference;
    property PhaseOffset: Real read FPhaseOffset write SetPhaseOffset;
    property Prebuffer: Real read GetPrebuffer write SetPrebuffer;
    property OnParamsChange: TSGParamsEvent read FOnParamsChange write FOnParamsChange;
  end;

const
  SGFrequencyRange: TRange = (Min: 1.e-3; Max: 1.e5);
  SGLevelRange: TRange = (Min: -114.0; Max: 0.0);

implementation

uses Math;

function AmpOf(const Level: Real; ResBits: Integer): Real;
begin
  Result := dBToLin(Level) * (1 shl (ResBits - 1) - 1);
end;

function PhaseIncOf(const Frequency: Real; SampleRate: Integer): Real;
begin
  Result := _2Pi * Frequency / SampleRate;
end;

{ TSineGen }

constructor TSineGen.Create;
begin
  inherited Create;
  FFrequency := 1000.0;
  FLevel_0 := -12.0;
  FLevel_1 := -12.0;
end;

function TSineGen.GetActive: Boolean;
begin
  Result := inherited Active;
end;

function TSineGen.GetDeviceID: Cardinal;
begin
  Result := inherited DeviceID;
end;

function TSineGen.GetLatency: Real;
begin
  Result := inherited Latency;
end;

function TSineGen.GetLevel: Real;
begin
  Result := MiddBLevel([FLevel_0, FLevel_1]);
end;

function TSineGen.GetMaxAvailFreq: Real;
begin
  Result := PCMFormat.SamplingRate / 2;
end;

function TSineGen.GetPCMFormat: TPCMFormat;
begin
  Result := inherited PCMFormat;
end;

function TSineGen.GetPrebuffer: Real;
begin
  Result := inherited Prebuffer;
end;

procedure TSineGen.ParamsChange(ParamMask: TSGParamMask);
begin
  if Assigned(FOnParamsChange) then
    FOnParamsChange(Self, ParamMask);
end;

procedure TSineGen.ProcessData(var Data; SampleCount: Integer);
type
  PInt24 = ^Int24;
  Int24 = array[0..2] of Byte;

  TByteArray = array[0..0] of Byte;
  TBytePairArray = array[0..0, 0..1] of Byte;

  TSmallIntArray = array[0..0] of SmallInt;
  TSmallIntPairArray = array[0..0, 0..1] of SmallInt;

  TInt24Array = array[0..0] of Int24;
  TInt24PairArray = array[0..0, 0..1] of Int24;

  TIntegerArray = array[0..0] of Integer;
  TIntegerPairArray = array[0..0, 0..1] of Integer;
var
  HalfPhaseDif: Real;
  PhaseInc_Sh, Amp_0_Sh, Amp_1_Sh, PhaseOfs_0_Sh, PhaseOfs_1_Sh: Extended;
  FormatTag, I, Value_0, Value_1: Integer;
begin
  if SampleCount <= 0 then
    Exit;
  with FWaveState, PCMFormat do
  begin
    FormatTag := BitDepth div 4 + ChannelCount - 3;
    if ChannelCount > 1 then
      HalfPhaseDif := FPhaseDifference / 2.0
    else
      HalfPhaseDif := 0.0;
    PhaseInc_Sh := (PhaseIncOf(FFrequency, SamplingRate) - PhaseInc) / SampleCount;
    Amp_0_Sh := (AmpOf(FLevel_0, BitDepth) - Amp_0) / SampleCount;
    Amp_1_Sh := (AmpOf(FLevel_1, BitDepth) - Amp_1) / SampleCount;
    PhaseOfs_0_Sh := (FPhaseOffset - HalfPhaseDif - PhaseOfs_0) / SampleCount;
    PhaseOfs_1_Sh := (FPhaseOffset + HalfPhaseDif - PhaseOfs_1) / SampleCount;
    for I := 0 to SampleCount - 1 do
    begin
      Value_0 := Round(Amp_0 * Sin(PhaseOfs_0 + Phase));
      if ChannelCount > 1 then
        Value_1 := Round(Amp_1 * Sin(PhaseOfs_1 + Phase));
      case FormatTag of
        0: TByteArray(Data)[I] := 128 + Value_0;
        1:
          begin
            TBytePairArray(Data)[I, 0] := 128 + Value_0;
            TBytePairArray(Data)[I, 1] := 128 + Value_1;
          end;
        2: TSmallIntArray(Data)[I] := Value_0;
        3:
          begin
            TSmallIntPairArray(Data)[I, 0] := Value_0;
            TSmallIntPairArray(Data)[I, 1] := Value_1;
          end;
        4: TInt24Array(Data)[I] := PInt24(@Value_0)^;
        5:
          begin
            TInt24PairArray(Data)[I, 0] := PInt24(@Value_0)^;
            TInt24PairArray(Data)[I, 1] := PInt24(@Value_1)^;
          end;
        6: TIntegerArray(Data)[I] := Value_0;
        7:
          begin
            TIntegerPairArray(Data)[I, 0] := Value_0;
            TIntegerPairArray(Data)[I, 1] := Value_1;
          end;
      end;
      Phase := Phase + PhaseInc;
      if Phase > _2Pi then
        Phase := Phase - _2Pi;
      PhaseInc := PhaseInc + PhaseInc_Sh;
      Amp_0 := Amp_0 + Amp_0_Sh;
      Amp_1 := Amp_1 + Amp_1_Sh;
      PhaseOfs_0 := PhaseOfs_0 + PhaseOfs_0_Sh;
      PhaseOfs_1 := PhaseOfs_1 + PhaseOfs_1_Sh;
    end;
  end;
end;

procedure TSineGen.RestoreWaveState(const Source);
begin
  FWaveState := TSGWaveState(Source);
end;

procedure TSineGen.SaveWaveState(var Dest);
begin
  TSGWaveState(Dest) := FWaveState;
end;

procedure TSineGen.SetActive(const Value: Boolean);
begin
  if Value <> Active then
  begin
    if Value then
    begin
      SetFrequency(Min(FFrequency, MaxAvailFreq));
      if (PCMFormat.ChannelCount < 2) and (FLevel_0 <> FLevel_1) then
        SetLevel(RoundTo(GetLevel, -2));
      FillChar(FWaveState, SizeOf(TSGWaveState), 0);
      FWaveState.PhaseInc := PhaseIncOf(FFrequency, PCMFormat.SamplingRate);
    end;
    FPhaseOffset := 0.0;
    inherited Active := Value;
    ParamsChange([sgpActive]);
  end;
end;

procedure TSineGen.SetDeviceID(const Value: Cardinal);
begin
  inherited DeviceID := Value;
  ParamsChange([sgpDevice]);
end;

procedure TSineGen.SetFrequency(const Value: Real);
var
  NewFreq: Real; 
begin
  if Active then
    NewFreq := EnsureRange(Value, SGFrequencyRange.Min, MaxAvailFreq)
  else
    NewFreq := EnsureRange(Value, SGFrequencyRange);
  if NewFreq <> FFrequency then
  begin
    FFrequency := NewFreq;
    UpdateStream;
    ParamsChange([sgpFrequency]);
  end;
end;

procedure TSineGen.SetLatency(const Value: Real);
begin
  inherited Latency := Value;
  ParamsChange([sgpLatency]);
end;

procedure TSineGen.SetLevel(const Value: Real);
var
  NewLevel: Real;
begin
  NewLevel := EnsureRange(Value, SGLevelRange);
  if (NewLevel <> FLevel_0) or (NewLevel <> FLevel_1) then
  begin
    FLevel_0 := NewLevel;
    FLevel_1 := NewLevel;
    UpdateStream;
    ParamsChange([sgpLevel, sgpLevel_0, sgpLevel_1]);
  end;
end;

procedure TSineGen.SetLevel_0(const Value: Real);
var
  NewLevel: Real;
begin
  NewLevel := EnsureRange(Value, SGLevelRange);
  if NewLevel <> FLevel_0 then
    if Stereo or not Active then
    begin
      FLevel_0 := NewLevel;
      UpdateStream;
      ParamsChange([sgpLevel, sgpLevel_0]);
    end else
      SetLevel(NewLevel);
end;

procedure TSineGen.SetLevel_1(const Value: Real);
var
  NewLevel: Real;
begin
  NewLevel := EnsureRange(Value, SGLevelRange);
  if NewLevel <> FLevel_1 then
    if Stereo or not Active then
    begin
      FLevel_1 := NewLevel;
      UpdateStream;
      ParamsChange([sgpLevel, sgpLevel_1]);
    end else
      SetLevel(NewLevel);
end;

procedure TSineGen.SetPCMFormat(const Value: TPCMFormat);
begin
  inherited PCMFormat := Value;
  ParamsChange([sgpPCMFormat]);
end;

procedure TSineGen.SetPhaseDifference(const Value: Real);
var
  NewPhaseDif: Real;
begin
  NewPhaseDif := EnsureRange(Value, -Pi, Pi);
  if NewPhaseDif <> FPhaseDifference then
  begin
    FPhaseDifference := NewPhaseDif;
    if Stereo then
      UpdateStream;
    ParamsChange([sgpPhaseDifference]);
  end;
end;

procedure TSineGen.SetPhaseOffset(const Value: Real);
begin
  if Active and (Value <> FPhaseOffset) then
  begin
    FPhaseOffset := Value;
    UpdateStream;
    ParamsChange([sgpPhaseOffset]);
  end;
end;

procedure TSineGen.SetPrebuffer(const Value: Real);
begin
  inherited Prebuffer := Value;
  ParamsChange([sgpPrebuffer]);
end;

function TSineGen.WaveStateDataSize: Integer;
begin
  Result := SizeOf(TSGWaveState);
end;

end.
