{-----------------|
| TWaveInOut v1.0 |
| (c) eCat        |
|-----------------}

unit WaveInOut;

interface

uses
  Windows, MMSystem, SysUtils, Classes, WaveFmt, GenUtils;

var
  DefPCMFormat: TPCMFormat = (
    BitDepth: 16;
    SamplingRate: 48000;
    ChannelCount: 2;
  );

const
  LatencyRange: TRange = (Min: 0.01; Max: 0.5);
  PrebufferRange: TRange = (Min: 0.1; Max: 5.0);

type
  EWaveError = class(Exception);
  
  TWaveInOut = class;
  TWaveStream = class;

  TWaveBuffer = class
  private
    FHeader: TWaveHdr;
    FNext: TWaveBuffer;
    FOwner: TWaveStream;
    FReturned: Boolean;
    procedure ProcessData;
    procedure Return;
  protected
    procedure AfterProcessData; virtual;
    procedure BeforeProcessData; virtual;
    function DataLength: Integer; virtual; abstract;
    procedure Process; virtual;
    procedure Send; virtual; abstract;
  public
    constructor Create(AOwner: TWaveStream); virtual;
    destructor Destroy; override;
    function WaveInOut: TWaveInOut;
    property Returned: Boolean read FReturned;
  end;

  TWaveBufferClass = class of TWaveBuffer;

  TWaveStream = class(TList)
  private
    FTerminated: Boolean;
    FEvent, FThread: THandle;
    FOwner: TWaveInOut;
    FReleasedCount: Integer;
    FFirst, FLast, FTail: TWaveBuffer;
    procedure Process;
    procedure SetLast(Value: TWaveBuffer);
  protected
    procedure Accept(Buffer: TWaveBuffer);
    function BufferClass: TWaveBufferClass; virtual; abstract;
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
    procedure Unload; virtual; abstract;
  public
    constructor Create(AOwner: TWaveInOut); virtual;
    destructor Destroy; override;
  end;

  TWaveStreamClass = class of TWaveStream;

  TWaveInOut = class
  private
    FActive: Boolean;
    FDeviceID: Cardinal;
    FPCMFormat: TPCMFormat;
    FStream: TWaveStream;
    FCaps: TWaveOutCaps;
    FPrebuffer: Real;
    FLatency: Real;
    FHandle: Integer;
    FFormat: TWaveFormatEx;
    function GetDeviceFormats: TStdPCMFormats;
    function GetDeviceName: string;
    function GetStereo: Boolean;
    procedure SetActive(const Value: Boolean);
    procedure SetDeviceID(const Value: Cardinal);
    procedure SetLatency(const Value: Real);
    procedure SetPCMFormat(const Value: TPCMFormat);
    procedure SetPrebuffer(const Value: Real);
  protected
    procedure CheckError(ErrorCode: Cardinal); virtual;
    procedure Close; virtual; abstract;
    procedure DoOpen(PFormatEx: PWaveFormatEx); virtual; abstract;
    procedure GetCaps; virtual; abstract;
    procedure Open; virtual;
    procedure ProcessData(var Data; SampleCount: Integer); virtual; abstract;
    function StreamClass: TWaveStreamClass; virtual; abstract;
    property Handle: Integer read FHandle;
  public
    constructor Create;
    destructor Destroy; override;
    property Active: Boolean read FActive write SetActive;
    property DeviceFormats: TStdPCMFormats read GetDeviceFormats;
    property DeviceID: Cardinal read FDeviceID write SetDeviceID;
    property DeviceName: string read GetDeviceName;
    property Latency: Real read FLatency write SetLatency;
    property PCMFormat: TPCMFormat read FPCMFormat write SetPCMFormat;
    property Prebuffer: Real read FPrebuffer write SetPrebuffer;
    property Stereo: Boolean read GetStereo;
  end;

  TWaveOutBuffer = class(TWaveBuffer)
  private
    FWaveEndState: Pointer;
  protected
    procedure AfterProcessData; override;
    procedure BeforeProcessData; override;
    function DataLength: Integer; override;
    procedure Process; override;
    procedure Send; override;
  public
    constructor Create(AOwner: TWaveStream); override;
    destructor Destroy; override;
  end;

  TWaveOutStream = class(TWaveStream)
  private
    FNeedUpdate: Boolean;
    procedure CheckUpdate;
  protected
    function BufferClass: TWaveBufferClass; override;
    procedure Unload; override;
  end;

  TWaveOut = class(TWaveInOut)
  protected
    procedure CheckError(ErrorCode: Cardinal); override;
    procedure Close; override;
    procedure DoOpen(PFormatEx: PWaveFormatEx); override;
    procedure GetCaps; override;
    procedure RestoreWaveState(const Source); virtual;
    procedure SaveWaveState(var Dest); virtual;
    function StreamClass: TWaveStreamClass; override;
    function WaveStateDataSize: Integer; virtual;
  public
    procedure UpdateStream;
  end;

implementation

uses Math;

procedure waveProc(Handle: Integer; Msg: Cardinal; Instance: TWaveInOut;
  Header: PWaveHdr; UnusedParam: Cardinal); stdcall;
begin
  if (Msg = WIM_DATA) or (Msg = WOM_DONE) then
    try
      TWaveBuffer(Header^.dwUser).Return;
    except
      if Assigned(ApplicationHandleException) then
        ApplicationHandleException(Instance);
    end;
end;

function WaveStreamProcess(Instance: TWaveStream): Cardinal; stdcall;
begin
  try
    Instance.Process;
  except
    if Assigned(ApplicationHandleException) then
      ApplicationHandleException(Instance);
  end;
  Result := 0;
end;

{ TWaveBuffer }

procedure TWaveBuffer.AfterProcessData;
begin
end;

procedure TWaveBuffer.BeforeProcessData;
begin
end;

constructor TWaveBuffer.Create(AOwner: TWaveStream);
begin
  FOwner := AOwner;
  with FHeader, WaveInOut do
  begin
    dwBufferLength := Round(FLatency * FFormat.nAvgBytesPerSec);
    GetMem(lpData, dwBufferLength);
    dwUser := Cardinal(Self);
  end;
end;

destructor TWaveBuffer.Destroy;
begin
  with FHeader do
    if lpData <> nil then
      FreeMem(lpData, dwBufferLength);
  inherited;
end;

procedure TWaveBuffer.Process;
begin
  if not FOwner.FTerminated then
    Send;
end;

procedure TWaveBuffer.ProcessData;
begin
  BeforeProcessData;
  with FHeader, WaveInOut do
    ProcessData(lpData^, DataLength div FFormat.nBlockAlign);
  AfterProcessData;
end;

procedure TWaveBuffer.Return;
begin
  FReturned := True;
  FOwner.Accept(Self);
end;

function TWaveBuffer.WaveInOut: TWaveInOut;
begin
  Result := FOwner.FOwner;
end;

{ TWaveStream }

procedure TWaveStream.Accept(Buffer: TWaveBuffer);
begin
  FTail := Buffer.FNext;
  Buffer.FNext := nil;
  SetLast(Buffer);
  if FFirst = nil then
  begin
    FFirst := Buffer;
    SetEvent(FEvent);
  end;
end;

constructor TWaveStream.Create(AOwner: TWaveInOut);
var
  I: Integer;
  ThreadID: Cardinal;
begin
  FOwner := AOwner;
  with FOwner do
    Capacity := Max(Round(FPrebuffer / FLatency), 2);
  for I := 1 to Capacity do
    Add(BufferClass.Create(Self));
  FEvent := CreateEvent(nil, False, False, nil);
  FThread := CreateThread(nil, 0, @WaveStreamProcess, Self, 0, ThreadID);
  SetThreadPriority(FThread, THREAD_PRIORITY_TIME_CRITICAL);
  FFirst := First;
  SetEvent(FEvent);
end;

destructor TWaveStream.Destroy;
begin
  FTerminated := True;
  Unload;
  WaitForSingleObject(FThread, INFINITE);
  CloseHandle(FThread);
  CloseHandle(FEvent);
  inherited Destroy;
end;

procedure TWaveStream.Notify(Ptr: Pointer; Action: TListNotification);
begin
  case Action of
    lnAdded: SetLast(Ptr);
    lnDeleted: TObject(Ptr).Free;
  end;
end;

procedure TWaveStream.Process;
begin
  while WaitForSingleObject(FEvent, INFINITE) = WAIT_OBJECT_0 do
    while FFirst <> nil do
    begin
      FFirst.Process;
      if FTerminated then
      begin
        Inc(FReleasedCount);
        if FReleasedCount = Count then
          Exit;
      end;
      FFirst := FFirst.FNext;
    end;
  RaiseLastOSError;
end;

procedure TWaveStream.SetLast(Value: TWaveBuffer);
begin
  if FLast <> nil then
    FLast.FNext := Value;
  FLast := Value;
end;

{ TWaveInOut }

procedure TWaveInOut.CheckError(ErrorCode: Cardinal);
begin
end;

constructor TWaveInOut.Create;
begin
  FLatency := 0.05;
  FPrebuffer := 0.1;
  FPCMFormat := DefPCMFormat;
  SetDeviceID(WAVE_MAPPER);
end;

destructor TWaveInOut.Destroy;
begin
  SetActive(False);
  inherited;
end;

function TWaveInOut.GetDeviceFormats: TStdPCMFormats;
begin
  Result := TStdPCMFormats(FCaps.dwFormats);
end;

function TWaveInOut.GetDeviceName: string;
begin
  Result := FCaps.szPname;
end;

function TWaveInOut.GetStereo: Boolean;
begin
  Result := FPCMFormat.ChannelCount > 1;
end;

procedure TWaveInOut.Open;
var
  wfe: TWaveFormatExtensible;
begin
  SetPCMFormatInfo(wfe, FPCMFormat);
  DoOpen(@wfe.Format);
  FFormat := wfe.Format;
end;

procedure TWaveInOut.SetActive(const Value: Boolean);
begin
  if Value <> FActive then
  begin
    if Value then
    begin        
      Open;
      try
        FStream := StreamClass.Create(Self);
      except
        Close;
        raise;
      end;
    end
    else
    begin
      FStream.Free;
      Close;
    end;
    FActive := Value;
  end;
end;

procedure TWaveInOut.SetDeviceID(const Value: Cardinal);
begin
  if not FActive then
  begin
    FDeviceID := Value;
    FillChar(FCaps, SizeOf(TWaveOutCaps), 0);
    GetCaps;
  end;
end;

procedure TWaveInOut.SetLatency(const Value: Real);
begin
  FLatency := EnsureRange(Value, LatencyRange);
end;

procedure TWaveInOut.SetPCMFormat(const Value: TPCMFormat);
begin
  if not FActive then
    FPCMFormat := ValidPCMFormat(Value);
end;

procedure TWaveInOut.SetPrebuffer(const Value: Real);
begin
  FPrebuffer := EnsureRange(Value, PrebufferRange);
end;

{ TWaveOutBuffer }

procedure TWaveOutBuffer.AfterProcessData;
begin
  TWaveOut(WaveInOut).SaveWaveState(FWaveEndState^);
end;

procedure TWaveOutBuffer.BeforeProcessData;
begin
  TWaveOutStream(FOwner).CheckUpdate;
end;

constructor TWaveOutBuffer.Create(AOwner: TWaveStream);
begin
  inherited Create(AOwner);
  WaveInOut.CheckError(waveOutPrepareHeader(WaveInOut.FHandle, @FHeader,
    SizeOf(TWaveHdr)));
  GetMem(FWaveEndState, TWaveOut(WaveInOut).WaveStateDataSize);
end;

function TWaveOutBuffer.DataLength: Integer;
begin
  Result := FHeader.dwBufferLength;
end;

destructor TWaveOutBuffer.Destroy;
begin
  if FWaveEndState <> nil then
    FreeMem(FWaveEndState);
  waveOutUnprepareHeader(WaveInOut.FHandle, @FHeader, SizeOf(TWaveHdr));
  inherited Destroy;
end;

procedure TWaveOutBuffer.Process;
begin
  if not FOwner.FTerminated then
    ProcessData;
  inherited Process;
end;

procedure TWaveOutBuffer.Send;
begin
  waveOutWrite(WaveInOut.FHandle, @FHeader, SizeOf(TWaveHdr));
end;

{ TWaveOutStream }

function TWaveOutStream.BufferClass: TWaveBufferClass;
begin
  Result := TWaveOutBuffer;
end;

procedure TWaveOutStream.CheckUpdate;
var
  Buf: TWaveBuffer;
begin
  if FNeedUpdate then
  begin
    FNeedUpdate := False;
    Buf := FTail;
    if Buf <> nil then
    begin
      TWaveOut(FOwner).RestoreWaveState(TWaveOutBuffer(Buf).FWaveEndState^);
      Buf := Buf.FNext;
      while Buf <> FFirst do
      begin
        Buf.ProcessData;
        Buf := Buf.FNext;
      end;
    end;
  end;
end;

procedure TWaveOutStream.Unload;
begin
  waveOutReset(FOwner.FHandle);
end;

{ TWaveOut }

procedure TWaveOut.CheckError(ErrorCode: Cardinal);
var
  Buf: array[0..1023] of Char;
begin
  if ErrorCode <> MMSYSERR_NOERROR then
  begin
    waveOutGetErrorText(ErrorCode, Buf, SizeOf(Buf));
    raise EWaveError.Create(Buf);
  end;
end;

procedure TWaveOut.Close;
begin
  waveOutClose(FHandle);
end;

procedure TWaveOut.GetCaps;
begin
  waveOutGetDevCaps(FDeviceID, @FCaps, SizeOf(TWaveOutCaps));
end;

procedure TWaveOut.DoOpen(PFormatEx: PWaveFormatEx);
begin
  CheckError(waveOutOpen(@FHandle, FDeviceID, PFormatEx, Cardinal(@waveProc),
    Cardinal(Self), WAVE_FORMAT_DIRECT * Ord(Win32MajorVersion > 4) or
    CALLBACK_FUNCTION));
end;

procedure TWaveOut.RestoreWaveState(const Source);
begin
end;

procedure TWaveOut.SaveWaveState(var Dest);
begin
end;

function TWaveOut.StreamClass: TWaveStreamClass;
begin
  Result := TWaveOutStream;
end;

procedure TWaveOut.UpdateStream;
begin
  if FActive then
    TWaveOutStream(FStream).FNeedUpdate := True;
end;

function TWaveOut.WaveStateDataSize: Integer;
begin
  Result := 0;
end;

end.
